#ifndef SDP4_h
#define SDP4_h

/*+
 * The SDP4 class is a base class to calculate ephemeris for
 * an artificial Earth satellite, using the SGP4 and SDP4 models.
 *
 * The SGP4 and SDP4 models of satellite prediction
 * (Felix R. Hoots, Roland L. Roehrich, T.S. Kelso, 1980, 1988,
 * Spacetrack report no. 3, Models for propagation of NORAD element sets,
 * http://www.celestrak.com/NORAD/documentation/)
 * give us algorithms to predict a few days' worth of motion of any
 * artificial satellite for which we have a NORAD Two Line Element set (TLE).
 * Hoots' report contains Fortran source code for three generations of
 * models: SGP, SGP4/SDP4 and SGP8/SDP8.  The TLE's are compiled assuming the
 * SGP4/SDP4 model, and that is the model that Kelso has ported to Pascal on
 * DOS.  (T.S. Kelso, 2000, TrakStar 2.64, http://www.celestrak.com,
 * T.S. Kelso, 1999, NORAD SGP4/SDP4 units 2.60, http://www.celestrak.com)
 * SGPn is used for near-Earth satellites (orbits shorter than 225 min)
 * while SDPn is used for satellites further afield.
 *
 * Hoots' Fortran code does not work under modern compilers, as it assumes
 * implicit zero initialisation of variables and persistence of local variables
 * between calls.  The code also uses GO TO statements rather than more
 * structured constructs like IF ELSE ENDIF blocks, which were not allowed
 * in Fortran at the time.  The SGPn models and the main parts of the SDPn
 * models are described mathematically in the report (albeit with a typo in
 * one SDPn equation).  But the most difficult code to port is the DEEP
 * subroutine, which is used by both SDP4 and SDP8.
 *
 * - In 1996 I translated the SGP8 and SDP8 models to C++.  This was done
 *   without having working Fortran code so that tests were limited to those
 *   shown in the Hoots et al. report.  These tests to not exercise the more
 *   complex parts of DEEP, namely e.g. satellites in a 12 hour resonance.
 *   That port is the code used in Sputnik up to version 1.9 (June 2002).
 * - In 2002 I further ported this C++ code to Java.  In the course of
 *   this work the test failure for satellites in resonance was discovered.
 *   This never went beyond the testing stage.
 * - As a result, a second port of SGP8/SDP8 from the Hoots et al. Fortran
 *   code to C++ and on to Java was made.  Testing was now extended in two
 *   ways.  Firstly, 18 satellites with between 0.1 and 16.5 revolutions per
 *   day are used.  Secondly, the reference data we want to match with the
 *   port are calculated with Kelso's Pascal port on DOS.  This port came
 *   close to reproducing Kelso's output for most satellites, but not for all.
 *   This also never went beyond the testing stage.
 * - In 2003 I started from scratch and ported the Hoots et al. code
 *   to Fortran 77, with some assistance from studying Kelso's Pascal code.
 *   - With minor modifications the code compiled under GNU g77, but did
 *     not give proper results.
 *   - Step by step more local variables were taken into common blocks to
 *     assure persistence and initialisation was finally moved to a separate
 *     subroutines where the variables are initialised by assignment rather
 *     than DATA.
 *   - The DEEP subroutine remained problematic due to its three ENTRYs
 *     points with different argument lists and its mixup between arguments and
 *     local variables.  Kelso's Pascal code pointed the way via three wrapper
 *     routines, an argument-free DEEP subroutine with three argument-free
 *     ENTRYs, and common blocks to pass arguments between wrappers and DEEP
 *     proper.
 *   - At this point the port reproduced Hoots' test data.  The
 *     18-satellite test against Kelso's application came close, but not close
 *     enough.  This later turned out to be a Y2K bug in THETAG.
 *   - At any rate the Fortran code was now prepared for porting to Java.
 *     Common variables would become state variables and subroutines and
 *     functions would become class methods.  The main problem was to change
 *     the flow control from GO TO to if/else blocks and for(;;) endless loops.
 *     In particular the DEEP routine was difficult to port; its structure had
 *     to be simplified by duplicating some code and in general despagging it.
 *   - The Java port then worked quite well.  After fixing the Y2K bug in
 *     THETAG it agrees within numeric precision with Kelso's application.
 *     The Fortran code does equally well, except for one satellite where after
 *     a long interval from the TLE epoch it is slightly less close to Kelso's
 *     application.
 *
 * TLEs can be picked up from http://www.celestrak.com/.
 * Current data are split into various groups of satellites.  Check
 * http://www.celestrak.com/NORAD/elements/master.shtml
 * to look up a satellite and which file it is in.  Historical elements are
 * also available.  http://www.wingar.demon.co.uk/satevo/
 * contains a few sets of TLEs including one very large set satbase.zip.
 * Note, however, that these use modified formats for the first of the three
 * lines: additional numeric parameters may be present before and after
 * column 22 of the line that is by standard reserved for the common name and
 * is by standard exactly 22 characters long.
 *
 * The position and velocity calculated here from the SGP4 and SDP4 models
 * is geocentric and (presumably) for the (mean?) equinox of date (the epoch
 * of the orbital elements).  It is unlikely to be for J2000 or B1950.
 *
 * To use this class:
 *
 * - Create an instance of the class.
 * - Invoke its Init() method.
 * - Invoke its NoradByName() method to read the TLE for one satellite
 *   from a file with one or more TLEs for different satellites.
 * - Invoke its GetPosVel() method with the Julian Date of interest.
 * - Retrieve some of the public state variables itsR[], itsV[],
 *   itsEpochJD, itsDesignator, itsName, itsNumber.
 * - Repeat GetPosVel() for different times.
 *
 * Copyright: (c) 2003 Horst Meyerdierks.
 *
 * This programme is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence as
 * published by the Free Software Foundation; either version 2 of
 * the Licence, or (at your option) any later version.
 *
 * This programme is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public Licence
 * along with this programme; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: SDP4.h,v 17.1 2003/04/06 08:42:31 hme Exp $
 *
 * $Log: SDP4.h,v $
 * Revision 17.1  2003/04/06 08:42:31  hme
 * Version 1.9.3.
 *
 * Revision 16.4  2003/04/05 19:38:57  hme
 * Sputnik 1.9 depends on the NoradNext method because it reads a whole
 * TLE file into a SatArray.  So have added that method.
 *
 * Revision 16.3  2003/04/05 18:50:38  hme
 * *** empty log message ***
 *
 * Revision 16.2  2003/04/05 17:55:57  hme
 * *** empty log message ***
 *
 * Author:
 *   Horst Meyerdierks, c/o Royal Observatory,
 *   Blackford Hill, Edinburgh, EH9 3HJ, Scotland;
 *   <hme@roe.ac.uk>
 *-*/

#include <stdio.h>

class SDP4 {

public:

  /** Position vector [Gm] */
  double *itsR;
  /** Velocity vector [km/s] */
  double *itsV;
  /** The name of the satellite. */
  char    itsName[23];
  /** Year and number of launch, plus part counter */
  char    itsDesignator[9];
  /** Number of the satellite from lines 1 or 2 */
  int     itsNumber;
  /** The TLE epoch expressed in JD minus 2450000 days. */
  double  itsEpochJD;

  int     itsEpochYear;     /* Year of epoch minus 1900 */
  double  itsEpochDay;      /* Julian day of epoch since start of year */

protected:

  /** Whether period is >= 225 min */
  int    itsIsDeep;


  double E1_XMO,E1_XNODEO,E1_OMEGAO,E1_EO,E1_XINCL,
    E1_XNO,E1_XNDT2O,E1_XNDD6O,E1_BSTAR,E1_X,E1_Y,E1_Z,
    E1_XDOT,E1_YDOT,E1_ZDOT,E1_EPOCH,E1_DS50;

  double C1_CK2,C1_CK4,C1_E6A,C1_QOMS2T,C1_S,C1_TOTHRD,
    C1_XJ3,C1_XKE,C1_XKMPER,C1_XMNPDA,C1_AE;

  double C2_DE2RA,C2_PI,C2_PIO2,C2_TWOPI,C2_X3PIO2;

  double SGP4_A1,SGP4_A3OVK2,SGP4_AO,SGP4_AODP,SGP4_AYCOF,
    SGP4_BETAO,SGP4_BETAO2,SGP4_C1,SGP4_C1SQ,
    SGP4_C2,SGP4_C3,SGP4_C4,SGP4_C5,SGP4_COEF,SGP4_COEF1,
    SGP4_COSIO,SGP4_D2,SGP4_D3,SGP4_D4,SGP4_DEL1,SGP4_DELMO,
    SGP4_DELO,SGP4_EETA,
    SGP4_EOSQ,SGP4_ETA,SGP4_ETASQ,SGP4_OMGCOF,SGP4_OMGDOT,
    SGP4_PERIGE,SGP4_PINVSQ,SGP4_PSISQ,
    SGP4_QOMS24,SGP4_S4,SGP4_SINIO,SGP4_SINMO,SGP4_T2COF,
    SGP4_T3COF,SGP4_T4COF,SGP4_T5COF,SGP4_TEMP,SGP4_TEMP1,
    SGP4_TEMP2,SGP4_TEMP3,SGP4_THETA2,SGP4_THETA4,SGP4_TSI,
    SGP4_X1M5TH,SGP4_X1MTH2,SGP4_X3THM1,SGP4_X7THM1,
    SGP4_XHDOT1,SGP4_XLCOF,SGP4_XMCOF,SGP4_XMDOT,SGP4_XNODCF,
    SGP4_XNODOT,SGP4_XNODP;
  int SGP4_ISIMP;

  double SDP4_A1,SDP4_A3OVK2,SDP4_AO,SDP4_AODP,
    SDP4_AYCOF,SDP4_BETAO,SDP4_BETAO2,SDP4_C1,SDP4_C2,
    SDP4_C4,SDP4_COEF,SDP4_COEF1,SDP4_COSG,SDP4_COSIO,
    SDP4_DEL1,SDP4_DELO,SDP4_EETA,SDP4_EOSQ,
    SDP4_ETA,SDP4_ETASQ,SDP4_OMGDOT,SDP4_PERIGE,SDP4_PINVSQ,
    SDP4_PSISQ,SDP4_QOMS24,SDP4_S4,SDP4_SING,
    SDP4_SINIO,SDP4_T2COF,SDP4_TEMP1,SDP4_TEMP2,SDP4_TEMP3,
    SDP4_THETA2,SDP4_THETA4,SDP4_TSI,SDP4_X1M5TH,
    SDP4_X1MTH2,SDP4_X3THM1,SDP4_X7THM1,SDP4_XHDOT1,SDP4_XLCOF,
    SDP4_XMDOT,SDP4_XNODCF,SDP4_XNODOT,SDP4_XNODP;

  double DEEP_A1,DEEP_A2,DEEP_A3,DEEP_A4,DEEP_A5,DEEP_A6,
    DEEP_A7,DEEP_A8,DEEP_A9,DEEP_A10,DEEP_AINV2,DEEP_ALFDP,
    DEEP_AQNV,DEEP_ATIME,DEEP_BETDP,DEEP_BFACT,DEEP_C,DEEP_CC,
    DEEP_COSIS,DEEP_COSOK,DEEP_COSQ,DEEP_CTEM,DEEP_D2201,
    DEEP_D2211,DEEP_D3210,DEEP_D3222,DEEP_D4410,DEEP_D4422,
    DEEP_D5220,DEEP_D5232,DEEP_D5421,DEEP_D5433,DEEP_DALF,
    DEEP_DAY,DEEP_DBET,DEEP_DEL1,DEEP_DEL2,DEEP_DEL3,DEEP_DELT,
    DEEP_DLS,DEEP_E3,DEEP_EE2,DEEP_EOC,DEEP_EQ,DEEP_F2,
    DEEP_F220,DEEP_F221,DEEP_F3,DEEP_F311,DEEP_F321,DEEP_F322,
    DEEP_F330,DEEP_F441,DEEP_F442,DEEP_F522,DEEP_F523,
    DEEP_F542,DEEP_F543,DEEP_FASX2,DEEP_FASX4,DEEP_FASX6,
    DEEP_FT,DEEP_G200,DEEP_G201,DEEP_G211,DEEP_G300,DEEP_G310,
    DEEP_G322,DEEP_G410,DEEP_G422,DEEP_G520,DEEP_G521,DEEP_G532,
    DEEP_G533,DEEP_GAM,DEEP_OMEGAQ,DEEP_PE,DEEP_PGH,DEEP_PH,
    DEEP_PINC,DEEP_PL,DEEP_PREEP,DEEP_S1,DEEP_S2,
    DEEP_S3,DEEP_S4,DEEP_S5,DEEP_S6,DEEP_S7,DEEP_SAVTSN,DEEP_SE,
    DEEP_SE2,DEEP_SE3,DEEP_SEL,DEEP_SES,DEEP_SGH,DEEP_SGH2,
    DEEP_SGH3,DEEP_SGH4,DEEP_SGHL,DEEP_SGHS,DEEP_SH,DEEP_SH2,
    DEEP_SH3,DEEP_SH1,DEEP_SHS,DEEP_SI,DEEP_SI2,DEEP_SI3,
    DEEP_SIL,DEEP_SINI2,DEEP_SINIS,DEEP_SINOK,DEEP_SINQ,
    DEEP_SINZF,DEEP_SIS,DEEP_SL,DEEP_SL2,DEEP_SL3,DEEP_SL4,
    DEEP_SLL,DEEP_SLS,DEEP_SSE,DEEP_SSG,DEEP_SSH,DEEP_SSI,
    DEEP_SSL,DEEP_STEM,DEEP_STEP2,DEEP_STEPN,DEEP_STEPP,
    DEEP_TEMP,DEEP_TEMP1,DEEP_THGR,DEEP_X1,DEEP_X2,DEEP_X2LI,
    DEEP_X2OMI,DEEP_X3,DEEP_X4,DEEP_X5,DEEP_X6,DEEP_X7,DEEP_X8,
    DEEP_XFACT,DEEP_XGH2,DEEP_XGH3,DEEP_XGH4,DEEP_XH2,DEEP_XH3,
    DEEP_XI2,DEEP_XI3,DEEP_XL,DEEP_XL2,DEEP_XL3,DEEP_XL4,
    DEEP_XLAMO,DEEP_XLDOT,DEEP_XLI,DEEP_XLS,
    DEEP_XMAO,DEEP_XNDDT,DEEP_XNDOT,DEEP_XNI,DEEP_XNO2,
    DEEP_XNODCE,DEEP_XNOI,DEEP_XNQ,DEEP_XOMI,DEEP_XPIDOT,
    DEEP_XQNCL,DEEP_Z1,DEEP_Z11,DEEP_Z12,DEEP_Z13,DEEP_Z2,
    DEEP_Z21,DEEP_Z22,DEEP_Z23,DEEP_Z3,DEEP_Z31,DEEP_Z32,
    DEEP_Z33,DEEP_ZCOSG,DEEP_ZCOSGL,DEEP_ZCOSH,DEEP_ZCOSHL,
    DEEP_ZCOSI,DEEP_ZCOSIL,DEEP_ZE,DEEP_ZF,DEEP_ZM,DEEP_ZMO,
    DEEP_ZMOL,DEEP_ZMOS,DEEP_ZN,DEEP_ZSING,DEEP_ZSINGL,
    DEEP_ZSINH,DEEP_ZSINHL,DEEP_ZSINI,DEEP_ZSINIL,DEEP_ZX,DEEP_ZY;
  int DEEP_IRESFL,DEEP_ISYNFL,DEEP_IRET,DEEP_IRETN,DEEP_LS;

  double DEEP_ZNS,DEEP_C1SS,DEEP_ZES,DEEP_ZNL,DEEP_C1L,
    DEEP_ZEL,DEEP_ZCOSIS,DEEP_ZSINIS,DEEP_ZSINGS,
    DEEP_ZCOSGS,DEEP_Q22,DEEP_Q31,DEEP_Q33,DEEP_G22,DEEP_G32,
    DEEP_G44,DEEP_G52,DEEP_G54,
    DEEP_ROOT22,DEEP_ROOT32,DEEP_ROOT44,DEEP_ROOT52,DEEP_ROOT54,
    DEEP_THDT;

  double DPINI_EQSQ,DPINI_SINIQ,DPINI_COSIQ,
    DPINI_RTEQSQ,DPINI_AO,DPINI_COSQ2,DPINI_SINOMO,DPINI_COSOMO,
    DPINI_BSQ,DPINI_XLLDOT,DPINI_OMGDT,DPINI_XNODOT,DPINI_XNODP;

  double DPSEC_XLL,DPSEC_OMGASM,DPSEC_XNODES,DPSEC_EM,
    DPSEC_XINC,DPSEC_XN,DPSEC_T;

public:

  /**
   * Initialise the SDP4.
   *
   * This initialises the SDP4 object.  Most state variables are set to
   * zero, some to constants necessary for the calculations. */

  void Init(void);

  /**
   * Read one named NORAD TLE from file.
   *
   * This method scans the named file for the given satellite name.  It
   * reads the two-line NORAD element set from the two lines behind the first
   * occurence of the satellite name.
   *
   * Although we look for two-line elements there are three lines per
   * satellite.  The format is:
   *
AAAAAAAAAAAAAAAAAAAAAA
1 NNNNNU NNNNNAAA NNNNN.NNNNNNNN +.NNNNNNNN +NNNNN-N +NNNNN-N N NNNNN
2 NNNNN NNN.NNNN NNN.NNNN NNNNNNN NNN.NNNN NNN.NNNN NN.NNNNNNNNNNNNNN
   *
   * The first line is 22 characters long and contains the common name of
   * the satellite.  This is the name to use when calling this method.  The
   * second line (first of the two TLE lines) contains the following
   * information:
   *
1 NNNNNU NNNNNAAA NNNNN.NNNNNNNN +.NNNNNNNN +NNNNN-N +NNNNN-N N NNNNN
  |---|                          satellite number
         |------|                international designator
         ||                      launch year last two digits in 1957-2056
           |-|                   launch number within year
              |-|                piece of launch
                  ||             year of epoch last two digits in 1957-2056
                    |----------| Julian Day of epoch
   1)                            |--------|
   2)                                       |------|
   3)                                                |------|
   ephemeris type                                             |
   element no.                                                  |--|
   check sum                                                        |
   *
   * - First time derivative of the mean motion or ballistic coefficient.
   * - Second time derivative of mean motion.  Field may be blank, meaning
   *   that the quantity is not available and must be assumed zero.  The format
   *   requires to read two integers and construct the number as the first
   *   divided by 100000 multiplied by 10 to the power of the second.
   * - BSTAR drag term if GP4 general perturbation theory was used.
   *   Otherwise, radiation pressure coefficient.
   *   The format requires to read two integers and construct the number
   *   as the first divided by 100000 multiplied by 10 to the power of the
   *   second.
   *
   * The third line (second of the two TLE lines) contains the following
   * information:
   *
2 NNNNN NNN.NNNN NNN.NNNN NNNNNNN NNN.NNNN NNN.NNNN NN.NNNNNNNNNNNNNN
  |---|                           satellite number
        |------|                  inclination [deg]
                 |------|         RA of ascending node [deg]
                          |-----| 1E7 times eccentricity
   argument of perigee [deg]      |------|
   mean anomaly [deg]                      |------|
   mean motion [rev/day]                            |---------|
   rev.no.                                                     |---|
   check sum                                                        |
   *
   * As a result of this routine the following public state variables are
   * set: itsName, itsDesignator, itsNumber, itsEpochJD.
   *
   * aFileName
   *   The name of the file that contains the TLE line triplets.
   * aName
   *   The name of the satellite, for comparison with the first line of the
   *   line triplets.  This name and the names read from file have to
   *   be equal in length and case-sensitive content, including any internal
   *   space characters. */

  void NoradByName(const char *aFileName, const char *aName, int *theStatus);

  /**
   * Read next NORAD TLE from file.
   * 
   * This method reads from the open file at its current position the next
   * set of two-line NORAD elements.  The file must be positioned in front
   * of the line with the satellite name, which is immediately followed by
   * the two lines of elements. */

  void NoradNext(FILE *aFile, int *theStatus);

  /**
   * Calculate position and velocity.
   *
   * This integrates the satellite orbit read previously with NoradByName
   * and thereby calculates the state variables itsR[] (in Gm) and itsV[]
   * (in km/s) for the time given.
   *
   * aJulDate
   *   The time for which the calculation should take place.  This must be
   *   given in units of days as the Julian Date minus 2450000 days. */

  void GetPosVel(double aJulDate);

  /**
   * Test routine.
   *
   * This test uses a spectrum of satellites with different periods
   * contained in a local file test.tle. */

  void driver2();

private:

  /**
   * Read TLE from open file.
   *
   * This method reads two lines at the current position of the open file,
   * and from these reads the satellite elements according to the NORAD
   * two-line format.  The file remains open and is left pointed behind the
   * two lines read.
   *
   * aFile
   *   The open file. */

  void ReadNorad12(FILE *aFile, int *theStatus);

  /**
   * Run the SGP4 model.
   *
   * This should be run for short-period satellites.  The criterion is
   * evaluated on reading the orbital data from the TLE and stored in the
   * state variable itsIsDeep (should be 0 for calling this routine).
   *
   * IFLAG
   *   IFLAG[0] must be given as 1 for the first call for any given satellite.
   *   It is then returned as 0 and can be given as 0 for further calls.
   * TSINCE
   *   TSINCE[0] is the time difference between the time of interest and the
   *   epoch of the TLE.  It must be given in minutes. */

  void RunSGP4(int *IFLAG, double *TSINCE);

  /**
   * Run the SDP4 model.
   *
   * This should be run for long-period satellites.  The criterion is
   * evaluated on reading the orbital data from the TLE and stored in the
   * state variable itsIsDeep (should be 1 for calling this routine).
   *
   * IFLAG
   *   IFLAG[0] must be given as 1 for the first call for any given satellite.
   *   It is then returned as 0 and can be given as 0 for further calls.
   * TSINCE
   *   TSINCE[0] is the time difference between the time of interest and the
   *   epoch of the TLE.  It must be given in minutes. */

  void RunSDP4(int *IFLAG, double *TSINCE);

  /**
   * Wrapper for deep space initialisation. */

  void DPINIT(double EOSQ, double SINIO, double COSIO,
    double BETAO, double AODP, double THETA2, double sing, double cosg,
    double BETAO2, double XMDOT, double OMGDOT, double XNODOTT, double XNODPP);

  /**
   * Wrapper for deep space secular effects. */

  void DPSEC(double *dpsec_args, double *TSINCE);

  /**
   * Wrapper for deep space lunar-solar periodics. */

  void DPPER(double *dpper_args);

  /**
   * Deep space initialisation. */

  void DEEP1(void);

  /**
   * Deep space secular effects. */

  void DEEP2(void);

  /**
   * Deep space lunar-solar periodics. */

  void DEEP3(void);

  double THETAG(double EP);
  double FMOD2P(double X);
  double ACTAN(double SINX, double COSX);
  char *hme_fgets(char *s, int n, FILE *stream);

};

#endif
